#######################################################################################################################################################################################
# 
# Note: N25Q512A83GSF40F
#
# 1. Basic Info
#    a. Sector Size       - 64KB (Subsector Size - 4KB, not supported by IP core)
#    b. Sector Erase Size - 64KB (Subsector Erase Size - 4KB, not supported by IP core)
#    c. Sector Number     - 1024 (0 ~ 1023)
#
# 2. Main Registers Info
#    a. RD_STATUS         - Status Register, contains the block protection bits that represent the memory sector to be protected from write or erase
#    b. RD_SID            - Only support in EPCS16 and EPCS64 flash device
#    c. RD_RDID
#       [31:8]:  reserved
#       [7:0]:   0x20     - Memory Capacity
#    d. MEM_OP
#       [31:24]: reserved
#       [24:8]:  Valid sector combination for sector protect and sector erase command (Important!!!!!!)
#       [7:2]:   reserved
#       [1:0]:   2'b11    - Sector protect: Active-high port that executes the sector protect operation.
#                         - If asserted, the IP takes the value of MEM_OP[23:8] and writes to the EPCQ status register.
#                         - The status register contains the block protection bits that represent the memory sector to be protected from write or erase.
#                2'b10    - Sector erase: Active-high port that executes the sector erase operation. 
#                         - If asserted, the IP starts erasing the memory sector on the EPCQ device based on MEM_OP[23:8] value.
#                2'b01    - Bulk erase: Active-high port that executes the bulk erase operation.
#                         - If asserted, the IP performs a full-erase operation that sets all memory bits of the EPCQ device to 1, 
#                         - which includes the general purpose memory of the EPCQ device.
#                         - ******Bulk erase is not supported in stack-die such as EPCQ512-L and EPCQ1024-L******
#                2'b00    - N/A, reserved
#
# 3. MEM_OP[23:8]: Valid sector combination for sector protect and sector erase command
#    a. Sector Protection - ******{MEM_OP[1:0] = 2'b11}******
#       MEM_OP[23:13]     - reserved
#       MEM_OP[12]        - TB  (Top/Bottom Bit)
#       MEM_OP[11]        - BP3 (Block Protection Bit, resort to datasheet page 14/15 for detail)
#       MEM_OP[10]        - BP2 (Block Protection Bit, resort to datasheet page 14/15 for detail)
#       MEM_OP[9]         - BP1 (Block Protection Bit, resort to datasheet page 14/15 for detail)
#       MEM_OP[8]         - BP0 (Block Protection Bit, resort to datasheet page 14/15 for detail)
#    b. Sector Erase      - ******{MEM_OP[1:0] = 2'b10}******
#       MEM_OP[23:10]     - reserved
#       MEM_OP[9:0]       - Sector Number 0~1023
#
# 4. Warning:
#    a. Register RD_SID (Offset 0x1) is invalid                                           - JTAG Avalon MM interface will run into crash once it try to access this register
#    b. Register MEM_OP (Offset 0x3) and Register CHIP_SELECT (Offset 0x6) are write only - Attribute: Write Only
#######################################################################################################################################################################################

#**********************************************************************************************************************
# Sub-function defination
#**********************************************************************************************************************
#1: target master service path

proc master_service_path_target { board_usb_port_no board_device_no } {

     set service_path_no {};
     
     #print master service paths array
     set jtag_master_array [get_service_paths master]
     format $jtag_master_array
     set jtag_master_array_length [llength $jtag_master_array]

     for {set i 0} {$i < $jtag_master_array_length} {incr i 1} {
          puts "jtag master #$i: [lindex $jtag_master_array $i]\n"
          if {[string match "/devices/10M50DA*@$board_device_no#USB-$board_usb_port_no*/JTAG/*master" [lindex $jtag_master_array $i]]} {
              set service_path_no $i
          }
     }

     #for {set i 0} {$i < $jtag_master_array_length} {incr i 1} {
     #     puts "jtag master #$i: [lindex $jtag_master_array $i]\n"
     #     if {[string match "/devices/10M08DA*@$board_device_no#USB-$board_usb_port_no*/JTAG/*master" [lindex $jtag_master_array $i]]} {
     #         set service_path_no $i
     #     }
     #}

     if {$service_path_no == ""} {
         error "No USB connection found"
     }

     #choose specific master service path
     set service_path [lindex $jtag_master_array $service_path_no]; # To choose correct master service path during test
     return $service_path
}

#----------------------------------------------------------------------------------------------------------------------
#2: master_write_x32
proc master_write_x32 { service_path master_addr data_write_array} {

     format $data_write_array
     set data_write_length [llength $data_write_array]

     for {set i 0} {$i < $data_write_length} {incr i 1} {
         master_write_32 $service_path [format 0x%x [expr $master_addr + [expr $i * 4]]] [lindex $data_write_array $i]
         #puts "Current write address: [format 0x%x [expr $master_addr + [expr $i * 4]]]\n"
         }
}

#----------------------------------------------------------------------------------------------------------------------
#3: master_read_x32
proc master_read_x32 { service_path master_addr data_read_length } {

     set master_read_array [master_read_32 $service_path $master_addr $data_read_length]
     format $master_read_array
     return $master_read_array
}

#----------------------------------------------------------------------------------------------------------------------
#4: qspi_write_x32
proc qspi_write_x32 { service_path master_addr offset_addr data_write_array} {

     format $data_write_array
     set data_write_length [llength $data_write_array]

     for {set i 0} {$i < $data_write_length} {incr i 1} {
         master_write_32 $service_path [format 0x%x [expr [format 0x%x [expr $master_addr + [expr $offset_addr * 4]]] + [expr $i * 4]]] [lindex $data_write_array $i]
         #puts "Current write address: [format 0x%x [expr [format 0x%x [expr $master_addr + [expr $offset_addr * 4]]] + [expr $i * 4]]]\n"
         }
}

#----------------------------------------------------------------------------------------------------------------------
#5: qspi_read_x32
proc qspi_read_x32 { service_path master_addr offset_addr data_read_length } {

     set master_read_array [master_read_32 $service_path [format 0x%x [expr $master_addr + [expr $offset_addr * 4]]] $data_read_length]
     format $master_read_array
     return $master_read_array
}

#**********************************************************************************************************************
# Main function
#**********************************************************************************************************************
set board_usb_port_no   1
set board_device_no     1

#----------------------------------------------------------------------------------------------------------------------
# test init
#----------------------------------------------------------------------------------------------------------------------

set result  0

#----------------------------------------------------------------------------------------------------------------------
#target correct master service path
set service_path [master_service_path_target $board_usb_port_no $board_device_no];
puts "current service path: $service_path\n"

#open master service
open_service master $service_path
puts "open service -> [is_service_open master $service_path]\n"

# #**********************************************************************************************************************
# #Product Info
# #**********************************************************************************************************************
# set prod_info_master_addr   0x00000000; # Qsys JTAG-Avalon Memory Map Address: 0x00000000 ~ 0x0000000F
# set prod_info_read_num                8
# 
# puts "/////////////////////////////////////////////////////////////////////////////////////////////////////////////////"
# puts "Product Info Checkout......"
# puts "/////////////////////////////////////////////////////////////////////////////////////////////////////////////////\n"
# set read_array [master_read_x32 $service_path $prod_info_master_addr $prod_info_read_num]
# format $read_array
# puts "Address:  $prod_info_master_addr\n"
# puts "Sequence: $read_array\n"

#**********************************************************************************************************************
#QSPI Access
#**********************************************************************************************************************
#EPCQ Serial Flash Controller Register Offset
set epcq_rd_status_offset   0x0; # R, Status Register, contains the block protection bits that represent the memory sector to be protected from write or erase
set epcq_rd_sid_offset      0x1; # R, invalid here as JTAG Avalon MM interface will run into crash once it try to access this register
set epcq_rd_rdid_offset     0x2; # R
set epcq_mem_op_offset      0x3; # Write Only
set epcq_isr_offset         0x4; # WR
set epcq_imr_offset         0x5; # WR
set epcq_chip_select_offset 0x6; # Write Only

#EPCQ Serial Flash Controller Qsys Address
set qspi_csr_master_addr    0x10000000; # Qsys JTAG-Avalon Memory Map Address: 0x10000000 ~ 0x1000001F
set qspi_mem_master_addr    0x14000000; # Qsys JTAG-Avalon Memory Map Address: 0x14000000 ~ 0x17FFFFFF
#set qspi_csr_master_addr   0x04000000; # Qsys JTAG-Avalon Memory Map Address: 0x04000000 ~ 0x0400001F
#set qspi_mem_master_addr   0x00000000; # Qsys JTAG-Avalon Memory Map Address: 0x00000000 ~ 0x03FFFFFF
set qspi_mem_data_offset    0x0
set qspi_mem_data_array     "0x00112233 0x44556677 0x8899aabb 0xccddeeff"

#Other parameters definition
set cmd_read_num            1
set data_read_num           4

puts "/////////////////////////////////////////////////////////////////////////////////////////////////////////////////"
puts "QSPI Access Checkout......"
puts "/////////////////////////////////////////////////////////////////////////////////////////////////////////////////\n"
#Memory Capacity Info
set read_array [qspi_read_x32 $service_path $qspi_csr_master_addr $epcq_rd_rdid_offset $cmd_read_num]
format $read_array
puts "EPCQ_RD_RDID: Memory Capacity (0x20)\n"
puts "Address:  [format 0x%x [expr $qspi_csr_master_addr + [expr $epcq_rd_rdid_offset * 4]]]\n"
puts "Sequence: $read_array\n"

#Interrupt Mask Checkout
set read_array [qspi_read_x32 $service_path $qspi_csr_master_addr $epcq_imr_offset $cmd_read_num]
format $read_array
puts "EPCQ_IMR: Mask Of Interrupt Status Register\n"
puts "Address:  [format 0x%x [expr $qspi_csr_master_addr + [expr $epcq_imr_offset * 4]]]\n"
puts "Sequence: $read_array\n"

#----------------------------------------------------------------------------------------------------------------------
#Sector #0 Checkout
#----------------------------------------------------------------------------------------------------------------------
#Clear Interrupt
qspi_write_x32 $service_path $qspi_csr_master_addr $epcq_isr_offset 0x00000000

#Remove protection for all sectors
qspi_write_x32 $service_path $qspi_csr_master_addr $epcq_mem_op_offset 0x00001003; #0x00000003

#Status Register Info
set read_array [qspi_read_x32 $service_path $qspi_csr_master_addr $epcq_rd_status_offset $cmd_read_num]
format $read_array
puts "EPCQ_RD_STATUS: Read Status Register\n"
puts "Address:  [format 0x%x [expr $qspi_csr_master_addr + [expr $epcq_rd_status_offset * 4]]]\n"
puts "Sequence: $read_array\n"

#Erase Sector #0
qspi_write_x32 $service_path $qspi_csr_master_addr $epcq_mem_op_offset 0x00000002
after 2000;

#Status Register Info
set read_array [qspi_read_x32 $service_path $qspi_csr_master_addr $epcq_rd_status_offset $cmd_read_num]
format $read_array
puts "EPCQ_RD_STATUS: Read Status Register\n"
puts "Address:  [format 0x%x [expr $qspi_csr_master_addr + [expr $epcq_rd_status_offset * 4]]]\n"
puts "Sequence: $read_array\n"

#Write a 32-bit data to Sector #0
format $qspi_mem_data_array
qspi_write_x32 $service_path $qspi_mem_master_addr $qspi_mem_data_offset $qspi_mem_data_array
after 2000;

#Status Register Info
set read_array [qspi_read_x32 $service_path $qspi_csr_master_addr $epcq_rd_status_offset $cmd_read_num]
format $read_array
puts "EPCQ_RD_STATUS: Read Status Register\n"
puts "Address:  [format 0x%x [expr $qspi_csr_master_addr + [expr $epcq_rd_status_offset * 4]]]\n"
puts "Sequence: $read_array\n"

#Read a 32-bit data from Sector #0
set read_array [qspi_read_x32 $service_path $qspi_mem_master_addr $qspi_mem_data_offset $data_read_num]
format $read_array
puts "EPCQ Sector #0 Read/Write Checkout\n"
puts "Address:  [format 0x%x [expr $qspi_mem_master_addr + [expr $qspi_mem_data_offset * 4]]]\n"
puts "Sequence: $read_array\n"

#Erase Sector #0
qspi_write_x32 $service_path $qspi_csr_master_addr $epcq_mem_op_offset 0x00000002
after 2000;

#Read a 32-bit data from Sector #0
set read_array [qspi_read_x32 $service_path $qspi_mem_master_addr $qspi_mem_data_offset $data_read_num]
format $read_array
puts "EPCQ Sector #0 Read/Write Checkout\n"
puts "Address:  [format 0x%x [expr $qspi_mem_master_addr + [expr $qspi_mem_data_offset * 4]]]\n"
puts "Sequence: $read_array\n"

#close master service
close_service master $service_path
puts "close service -> [is_service_open master $service_path]\n"
#----------------------------------------------------------------------------------------------------------------------
#result
return $result
######################################################################################################################################################
######################################################################################################################################################